#include "gebs/gebs.h"
#include "mongoose/mongoose.h"
#include "cJSON/cJSON.h"
#include "md5-c/md5.h"

#include "routes.h"
#include "baked.h"
#include "clonestr.h"

#define INTERNAL_SERVER_ERROR_MSG "Internal server error ;(. Try again later..."

void make_internal_server_error(Route_Result *result)
{
    result->status_code = 500;
    list_append(&result->headers, "Content-Type: text/plain");
    sb_append_nstr(&result->body, INTERNAL_SERVER_ERROR_MSG);
    sb_finish(&result->body);
}

void make_application_json(Route_Result *result, int code, cJSON *root)
{
    char *root_text = cJSON_PrintUnformatted(root);
    defer { free(root_text); }

    result->status_code = code;
    list_append(&result->headers, clonestr("Content-Type: application/json"));
    sb_append_nstr(&result->body, root_text);
    sb_finish(&result->body);
}

void make_image_xicon(Route_Result *result, int code, char *icon)
{
    result->status_code = code;
    list_append(&result->headers, clonestr("Content-Type: image/x-icon"));
    sb_append_nstr(&result->body, icon);
    sb_finish(&result->body);
}

void make_text(Route_Result *result, char *subtype, int code, char *in)
{
    char type[100];
    snprintf(type, sizeof(type), "Content-Type: text/%s", subtype);

    result->status_code = code;
    list_append(&result->headers, clonestr(type));
    sb_append_nstr(&result->body, in);
    sb_finish(&result->body);
}

bool gpp_run(char *path, NString_List *env, String_Builder *out)
{
    Cmd cmd = {0};
    defer { cmd_free(&cmd); }

    char gpp1[PATH_MAX];
    if (!get_baked_resource_path("gpp1", gpp1, sizeof(gpp1))) {
        return false;
    }

    cmd_append(&cmd, gpp1);
    cmd_append(&cmd, "-H");
    cmd_append(&cmd, "-x");
    cmd_append(&cmd, "--nostdinc");
    cmd_append(&cmd, path);

    for (size_t i = 0; i < env->count; i++) {
        cmd_append(&cmd, env->items[i]);
    }

    return cmd_run_collect(&cmd, out) == 0;
}

void route_build_id(struct mg_http_message *msg, Route_Result *result)
{
    cJSON *root = cJSON_CreateObject();
    defer { cJSON_Delete(root); }

    char *time = __TIME__;
    uchar md5_buf[16];
    md5String(time, md5_buf);
    String_Builder sb = {0};
    defer { sb_free(&sb); }
    for (size_t i = 0; i < 16; i++) {
        char tmp[3];
        snprintf(tmp, sizeof(tmp), "%02x", md5_buf[i]);
        sb_append_nstr(&sb, tmp);
    }
    sb_finish(&sb);

    cJSON_AddItemToObject(root, "build_id", cJSON_CreateString(sb.items));

    make_application_json(result, 200, root);
}

void route_page_not_found(struct mg_http_message *msg, Route_Result *result)
{
    NString_List env = {0};
    defer { list_free(&env); }

    list_append(&env, fmt("-DURL=%.*s", msg->uri.len, msg->uri.buf));

    String_Builder out = {0};
    defer { sb_free(&out); }
    
    char path[PATH_MAX] = {0};
    if (!get_baked_resource_path("page-missing.html", path, sizeof(path))) {
        make_internal_server_error(result);
        return;
    }
    
    #if DEBUG
        list_append(&env, "-DHOTRELOAD=1");
    #else
        list_append(&env, "-DHOTRELOAD=0");
    #endif

    bool ok = gpp_run(path, &env, &out);
    sb_finish(&out);

    if (!ok) {
        make_internal_server_error(result);
    } else {
        make_text(result, "html", 200, out.items);
    }
}

void route_simple_css(struct mg_http_message *msg, Route_Result *result)
{
    char path[PATH_MAX] = {0};
    if (!get_baked_resource_path("simple.min.css", path, sizeof(path))) {
        make_internal_server_error(result);
        return;
    }

    String_Builder sb = {0};
    defer { sb_free(&sb); }
    if (!sb_read_file(&sb, path)) {
        make_internal_server_error(result);
        return;
    }
    sb_finish(&sb);

    make_text(result, "css", 200, sb.items);
}

void route_favicon(struct mg_http_message *msg, Route_Result *result)
{
    char path[PATH_MAX] = {0};
    if (!get_baked_resource_path("favicon.ico", path, sizeof(path))) {
        make_internal_server_error(result);
        return;
    }

    String_Builder sb = {0};
    defer { sb_free(&sb); }
    if (!sb_read_file(&sb, path)) {
        make_internal_server_error(result);
        return;
    }
    sb_finish(&sb);

    make_image_xicon(result, 200, sb.items);
}

void route_hotreload_js(struct mg_http_message *msg, Route_Result *result)
{
    char path[PATH_MAX] = {0};
    if (!get_baked_resource_path("hotreload.js", path, sizeof(path))) {
        make_internal_server_error(result);
        return;
    }

    String_Builder sb = {0};
    defer { sb_free(&sb); }
    if (!sb_read_file(&sb, path)) {
        make_internal_server_error(result);
        return;
    }
    sb_finish(&sb);

    make_text(result, "javascript", 200, sb.items);
}

void route_home(struct mg_http_message *msg, Route_Result *result)
{
    NString_List env = {0};
    defer { list_free(&env); }

    String_Builder out = {0};
    defer { sb_free(&out); }

    char path[PATH_MAX] = {0};
    if (!get_baked_resource_path("home.html", path, sizeof(path))) {
        make_internal_server_error(result);
        return;
    }

    #if DEBUG
        list_append(&env, "-DHOTRELOAD=1");
    #else
        list_append(&env, "-DHOTRELOAD=0");
    #endif

    bool ok = gpp_run(path, &env, &out);
    sb_finish(&out);

    if (!ok) {
        make_internal_server_error(result);
    } else {
        make_text(result, "html", 200, out.items);
    }
}
